"""Functions to create experimental designs for portfolio choice models."""
# Experimental design functions

# Load modules
from locale import normalize
import numpy as np
import time
import datetime
import re
# from sklearn.preprocessing import MinMaxScaler
from pandas import get_dummies

# PortDesign class
class PortDesign:
    """Experimental design class

    This is the class to create experimental designs for portfolio choice
    models.

    Parameters
    ----------
    ATTLIST : list
        List that contains the specification of each alternative of the 
        experiment. Each element is a dictionary, in which each element 
        is an attribute. The keys of each element are the attribute names 
        and each value is a list that contains the attribute levels.
    NCS : int
        Number of choice situations.
    """
    def __init__(self, ATTLIST: list, NCS: int):
        # Define scalars
        self.ATTLIST = ATTLIST
        self.NCS = NCS
        self.NGOODS = len(self.ATTLIST)

        # Run integrity checks on the attribute list
        PortDesign.checkatts(ATTLIST,NCS)

        # Get variable names of the design matrix
        self.NAMELIST = []
        self.LEVS = []

        for  k in range(self.NGOODS):
            for key, value in self.ATTLIST[k].items():
                self.NAMELIST = self.NAMELIST  + [key + '_' + str(k+1)]
                self.LEVS = self.LEVS + [value]
        
        # Number of attributes
        self.NATT = len(self.NAMELIST)

    def generate_design(self,ITERLIM: int = None, NOIMPROVLIM: int = None, TIMELIM: float = None, CRIT: str = 'deff', COND: list = None, SEED: int = None, VERBOSE: bool = True):
        """Generate experimental design

        It generates an experimental design based on the specification 
        of the parent class and the parameters provided by the user.

        Parameters
        ----------
        ITERLIM : int, optional
            Iteration limit, by default None
        NOIMPROVLIM : int, optional
            Limit of iterations without improvement, by default None
        TIMELIM : float, optional
            Time limit in minutes, by default None
        CRIT : str, optional
            Efficiency criterion. If `maxcorr`, then the algorithm minimises 
            the maximum value of the correlation between attributes. If `deff`, 
            the algorithm maximises the D-efficiency of a linear model, by default 'deff'
        COND : list, optional
            Conditions between attributes of the design. See the examples for an overview
            , by default None
        SEED : int, optional
            Random seed, by default None
        VERBOSE : bool, optional
            Whether the algorithm prints output while optimising, by default True

        Returns
        -------
        bestdes : np.ndarray
            Optimal design
        perf : float
            Value of the efficiency criterion at the **first** iteration
        bestperf : float
            Value of the efficiency criterion at the **last** iteration
        elapsed_time : float
            Optimisation time
        best_t : int
            Number of iterations
        """
        # Set random seed if defined
        if SEED is not None:
            np.random.seed(SEED)

        # Generate conditions if defined
        if COND is not None:
            initconds = PortDesign.condgen('desmat',COND,self.NAMELIST,init=True)
            algconds = PortDesign.condgen('swapdes',COND,self.NAMELIST,init=False)
        else:
            initconds = None
            algconds = None

        # If conditions are not defined, set them to infinity
        if ITERLIM is not None:
            iterlim = ITERLIM
        else:
            iterlim = np.inf

        if NOIMPROVLIM is not None:
            noimprovlim = NOIMPROVLIM
        else:
            noimprovlim = np.inf

        if TIMELIM is not None:
            timelim = TIMELIM
        else:
            timelim = np.inf

        ############################################################
        ########## Step 1: Generate initial design matrix ##########
        ############################################################

        if VERBOSE:
            print('Generating the initial design matrix')

        desmat = PortDesign.initdesign(self.NAMELIST,self.LEVS,self.NATT,self.NGOODS,self.NCS,initconds)
        
        if CRIT == 'deff':
            raise ValueError("D-efficiency is not implemented yet")
            perf = PortDesign.deff(desmat,self.NATT,self.NGOODS,self.NCS)
        elif CRIT == 'maxcorr':
            perf = PortDesign.maxcorr(desmat,self.NATT,self.NGOODS,self.NCS)
        elif CRIT == 'maxcorr_wide':
            perf = PortDesign.maxcorr_wide(desmat,self.NATT,self.NGOODS,self.NCS)

        ############################################################
        ############## Step 2: Initialize algorighm ################
        ############################################################

        bestdes, bestperf, best_t, elapsed_time = PortDesign.swapalg(desmat,perf,self.NATT,self.NGOODS,self.NCS,algconds,CRIT,iterlim,noimprovlim,timelim)

        ############################################################
        ############## Step 3: Arange final design #################
        ############################################################

        # Reshape in long format
        bestdes = np.reshape(bestdes,(self.NCS*self.NGOODS,int(self.NATT/self.NGOODS)))

        # Add design and version rows
        grprow = np.repeat(np.arange(1,self.NCS+1),self.NGOODS)
        grprow.shape = (self.NCS*self.NGOODS,1)
        altrow = np.tile(np.arange(1,self.NGOODS+1),self.NCS)
        altrow.shape = (self.NCS*self.NGOODS,1)

        bestdes = np.c_[grprow,altrow,bestdes]

        # If verbose, then print output
        if VERBOSE:
            print('Optimization complete')
            print('Elapsed time: ' + str(datetime.timedelta(seconds=elapsed_time))[:7])
            print('Performance of initial design: ',round(perf,6))
            print('Performance of last stored design: ',round(bestperf,6))
            print('Algorithm iterations: ',best_t)
            print('')

        return bestdes, perf, bestperf, elapsed_time, best_t

    @staticmethod
    def checkatts(ATTLIST: list, NCS: int):
        
        NGOODS = len(ATTLIST)

        # Check if there are enough goods defined.
        assert NGOODS >= 2, "Error: at least two goods must be specified."

        # Start check loop among goods
        for k in range(NGOODS):

            # Check if the element k is a dictionary
            assert isinstance(ATTLIST[k],dict), "Error: element that defines good " + str(k+1) + " is not a dictionary."

            # Start check loop among elements of each good
            for key, value in ATTLIST[k].items():

                # Check if each element is a list
                assert isinstance(value, list), "Error: Attribute \'" + key + "\'" + " in good " + str(k+1) + " is not a list."

                # Check if NCS is divisible by number of attribute levels
                assert NCS%len(ATTLIST[k][key]) == 0, "Error: No. of Choice sets is not divisible by number of levels of attribute \'" + key + "\'" + " in good" + str(k+1) + "."

    @staticmethod
    def condgen(DESNAME: str, COND: list, NAMELIST: list, init: bool = False):
        
        # Match variable names with columns in the design matrix
        if init:
            DESCOLUMNS = [DESNAME + '[i,' + str(i) + ']' for i in range(len(NAMELIST))]
        else:
            DESCOLUMNS = [DESNAME + '[:,' + str(i) + ']' for i in range(len(NAMELIST))]

        # Create new list of conditions
        NEWCONDS = []

        for c in COND:
            # Take a copy of the string
            cc = c[:]

            # Replace names by design matrix columns
            for i in range(len(NAMELIST)):
                cc = cc.replace(NAMELIST[i],DESCOLUMNS[i])
            
            # If there is a conditional 'if' statement, then convert it to a logical 'or'
            if 'if' in cc:
                # Split the sting in the 'then' part
                if_part, then_part = cc.split('then')

                # Set the 'logical_not' operator in the 'if' part
                if_part = 'np.logical_not(' + if_part.replace('if','') + ')'

                # In case there are '&' in the 'then' part, convert them to '*'
                then_part = then_part.replace('&','*')

                # Merge if and then parts
                cc = 'np.logical_or(' + if_part + ',' + then_part + ')'
            
            # Finally, append condition to condition list
            NEWCONDS.append(cc)
        
        # ...and return the new conditions list
        return NEWCONDS

    @staticmethod
    def initdesign(NAMELIST: list, LEVS: list, NATT: int, NGOODS: int, NCS: int, COND: list):

        # Create and populate the initial design matrix
        desmat = []

        for k in range(NATT):
            col = np.repeat(LEVS[k],NCS/len(LEVS[k]))
            np.random.shuffle(col)
            desmat.append(col)
        
        desmat = np.array(desmat).T

        # Apply conditions if needed
        if COND is not None:
            for i in range(NCS):
                # Check if all conditions are satisfied. If not, do a big enough loop till all conditions are satisfied
                check_all = []

                for c in COND:
                    check_all.append(eval(c))

                check_all = np.all(check_all)

                if not check_all:
                    for _ in range(1000):
                        # Create a random vector of levels for the row in question
                        desmat[i,:] = np.array([np.random.choice(i) for i in LEVS])
                        
                        # Check if conditions are met with the new vector in the design.
                        check_all = []

                        for c in COND:
                            check_all.append(eval(c))

                        check_all = np.all(check_all)


                        # If so, break the loop and go for the next row
                        if check_all:
                            break
                
                # If after the big loop conditions are not met, then raise an error
                assert check_all, 'It is not possible to met all conditions in the initial design matrix.'
        
        # Return the design matrix
        return desmat

    @staticmethod
    def deff(DES,NATT,NGOODS,NCS):
        pass

    @staticmethod
    # Maximum correlation function
    def maxcorr(DES,NATT,NGOODS,NCS):
        X = np.reshape(DES,(NCS*NGOODS,int(NATT/NGOODS)))
        # X = DES.copy()
        cormat = np.corrcoef(X.T)
        maxcor = np.max(np.abs(cormat) - np.identity(len(cormat)))
        return(maxcor)

    @staticmethod
    # Maximum correlation function in wide form
    def maxcorr_wide(DES,NATT,NGOODS,NCS):
        # X = np.reshape(DES,(NCS*NGOODS,int(NATT/NGOODS)))
        X = DES.copy()
        cormat = np.corrcoef(X.T)
        maxcor = np.max(np.abs(cormat) - np.identity(len(cormat)))
        return(maxcor)

    @staticmethod
    def swapalg(DES,INITPERF,NATT,NGOODS,NCS,COND,CRIT,ITERLIM,NOIMPROVLIM,TIMELIM):
        
        # Lock design matrix
        desmat = DES.copy()

        # Start stopwatch
        t0 = time.time()
        t1 = time.time()

        difftime = 0

        # Initialize algorithm parameters
        i = np.random.choice(np.arange(DES.shape[1]))
        t = 0
        ni = 0
        iterperf = INITPERF
        newperf = INITPERF

        # Start algorithm
        while True:
            
            # Iteration No.
            t = t+1
            
            # If one stopping criterion is satisfied, break!
            if ni >= NOIMPROVLIM or t >= ITERLIM or (difftime)/60 >= TIMELIM:
                break
            
            # Take a random swap
            pairswap = np.random.choice(desmat.shape[0],2,replace=False)

            # Check if attribute levels differ
            check_difflevels = desmat[pairswap[0],i] != desmat[pairswap[1],i]

            # If attribute levels differ, do the swap and check for conditions (if defined)
            if check_difflevels:
                swapdes = desmat.copy()
                swapdes[pairswap[0],i] = desmat[pairswap[1],i]
                swapdes[pairswap[1],i] = desmat[pairswap[0],i]
                
                # Check if conditions are satisfied after a swap
                check_all = []
                
                # If conditions are defined, this section will check that are satisfied, and rewrite 'check_satisfied_conds' if neccesary
                if COND is not None:
                    for c in COND:
                        check_all.append(eval(c))

                    check_all = np.all(check_all)
                
                # If all conditions are satisfied, compute D-error
                if check_all:
                    if CRIT == 'deff':
                        newperf = PortDesign.deff(swapdes,NATT,NGOODS,NCS)
                    elif CRIT == 'maxcorr':
                        newperf = PortDesign.maxcorr(swapdes,NATT,NGOODS,NCS)

            # ...else if they do not differ, keep the D-error
            else:
                newperf = iterperf.copy()
                
            # If the swap made an improvement, keep the design and update progress bar
            if CRIT == 'deff':
                improved = newperf > iterperf
            elif CRIT == 'maxcorr':
                improved = newperf < iterperf

            if improved:
                desmat = swapdes.copy()
                iterperf = newperf.copy()
                ni = 0
                
                # Update progress bar
                print('Optimizing. Press ESC to stop. / ' + 'Elapsed: ' + str(datetime.timedelta(seconds=difftime))[:7] + ' / Performance: ' + str(round(iterperf,6)),end='\r')
            
            # ...else, pass to a random attribute and increment the 'no improvement' counter by 1.
            else:
                i = np.random.choice(np.arange(DES.shape[1]))
                ni = ni+1
            
            # Update progress bar each second
            if (difftime)%1 < 0.1:
                print('Optimizing. Press ESC to stop. / ' + 'Elapsed: ' + str(datetime.timedelta(seconds=difftime))[:7] + ' / Performance: ' + str(round(iterperf,6)),end='\r',flush=True)
            
            t1 = time.time()
            difftime = t1-t0
        
        # Return optimal design plus efficiency
        return desmat, iterperf, t, difftime

    @staticmethod
    def orthcode(X):
        pass


"""BFGSMin functions"""
# BFGSMin functions
# Written by Jose Ignacio Hernandez

# Import packages
import numpy as np

# BFGS Minimizer function
def _bfgsmin(f,x0,maxiter=1000,tol=np.sqrt(np.finfo(float).eps),verbose=False,difftype='central',diffeps=np.sqrt(np.finfo(float).eps),steptol=1e-30,args=()):


    # Ignore warnings
    import warnings
    warnings.filterwarnings("ignore")

    # Initialize parameters
    x = np.array(x0)                            # Initial value of x
    f_val = f(x,*args)                          # Initial value for objective function
    g0 = _numgr(f,x,difftype,diffeps,*args)      # Initial value of gradient
    H0 = np.eye(x.shape[0])                     # Initial value of approximate Hessian is a positive definite matrix (e.g. identity matrix)
    g_diff = np.inf                             # Initial difference in gradient norm is set to infinity
    c1 = 1e-4                                   # Internal scalar for the Armijo-Goldstein condition (for step size computation)
    lambd = 1                                   # Initial value of the step size
    convergence = -1                            # Set convergence flag to -1. If zero, the algorithm converged

    # Print inital value of the objective function
    if verbose:
        print('Initial F-value: ' + str(round(f_val,2)))

    # Start algorithm
    for iter in range(maxiter):
        
        # If gradient norm is less than tolerance value, convergence is achieved and the loop is broken
        if g_diff < tol:
            convergence = 0

            if verbose:
                print('\nLocal minimum found. G-norm below tolerance')

            break

        # Set step size = 1 on each new iteration
        lambd = 1

        # Construct direction vector and relative gradient

        # d = np.linalg.solve(-H0,np.eye(len(x))) @ g0
        d = np.linalg.inv(-H0) @ g0
        m = d.T @ g0
        # Select step size that satisfies the Armijo-Goldstein condition
        while True:
            
            # If lambd decreases less than tol, then stop and return convergence = 3
            if lambd < steptol:
                if verbose:
                    print('\nLocal minimum possible. Step size tolerance limit reached.')
                convergence = 5
                break

            # Evaluate objective function using current step size
            x1 = x + lambd*d
            f1 = f(x1,*args)

            # Construct test for A-G condition
            ftest = f_val + c1*lambd*m

            # If A-G condition is satisfied (i.e current step size generates a decrease of the objective function), continue with BFGS algorithm
            if (np.isnan(f1) == False) and (f1 <= ftest) and (f1 > 0):
                break

            # ...else, decrease step size
            else:
                lambd = lambd/2

        # If step size calculation loop was broken, then break the algorithm and return convergence code + partial results
        if convergence == 5:
            break

        # BFGS ALGORITHM: construct the improvement and gradient improvement
        g1 = _numgr(f,x1,difftype,diffeps,*args)
        s0 = (lambd*d)[:,np.newaxis]
        y0 = (g1 - g0)[:,np.newaxis]

        # Update Hessian using BFGS formula
        H0 = H0 + (y0 @ y0.T) / (y0.T @ s0) - (((H0 @ s0) @ s0.T) @ H0)/((s0.T @ H0) @ s0)
        
        # Store new gradient and compute the new value of objective function
        x = x1.copy()
        g0 = g1.copy()
        f_val = f1.copy()

        # Compute infinite norm of gradient vector 
        g_diff = np.abs(m)
        # Print output
        if verbose:
            print('Iter No. ' + str(int(iter+1)) + ': F-value: ' + str(round(f_val,2)) + ' / Step size: ' + str(round(lambd,6)) + ' / G-norm: ' + str(round(g_diff,6)))

    # If the algorithm reaches the maximum iterations, it prints convergence message = 2
    if iter == maxiter:
        convergence = 2

    # Return the Hessian approximation
    H = H0

    # Return convergence flag, iterations, final f value, final x value, and final approx. hessian
    return({'convergence': convergence, 'iterations': iter+1, 'fun': f_val, 'x': x, 'hessian': H})

# Numeric gradient fuction
def _numgr(f,param,difftype='forward',eps=np.sqrt(np.finfo(float).eps),*args):    
    
    # Define scalars and initialize vectors
    K = len(param)                          # No. of parameters
    gr = np.full(K,np.nan)                  # Initialize gradient vector
    ej = np.eye(K)*eps                      # Vector of eps
    
    # If difftype == 'central', then:
    if difftype == 'central':
        for k in range(K):
            gr[k] = (f(param + ej[:,k],*args) - f(param - ej[:,k],*args))*0.5/eps

    # ...else, if difftype == 'forward' (default):
    elif difftype == 'forward':
        f0 = f(param,*args)
        for k in range(K):
            gr[k] = (f(param + ej[:,k],*args) - f0)/eps

    # ...else, return error
    else:
        raise ValueError('''difftype must be either 'forward' or 'central' ''')

    # Return gradient vector
    return(gr)


# Numeric Hessian function
class numhess:
    
    def __init__(self,f,eps=np.sqrt(np.finfo(float).eps)):
        self.f = f
        self.eps = eps
    
    def __call__(self,param,*args):
        eps = self.eps
        f = self.f

        # Define scalars and initialize vectors
        K = len(param)                          # No. of parameters
        hs = np.full((K,K),np.nan)              # Initialize Hessian vector
        ej = np.eye(K)*eps                      # Vector of eps

        # f0 = f(param,*args)
        for i in range(K):
            for j in range(K):
                # f1 = f(param + ej[:,i] + ej[:,j],*args)
                # f2 = f(param + ej[:,i],*args)
                # f3 = f(param + ej[:,j],*args)

                # hs[i,j] = (f1-f2-f3+f0)/(eps**2)
                # if i != j:
                #     hs[i,j] = (f1-f2-f3+f0)/(eps**2)
                
                f1 = f(param + ej[:,i] + ej[:,j],*args)
                f2 = f(param + ej[:,i] - ej[:,j],*args)
                f3 = f(param - ej[:,i] + ej[:,j],*args)
                f4 = f(param - ej[:,i] - ej[:,j],*args)

                hs[i,j] = (f1-f2-f3+f4)/(4*eps*eps)
                if i != j:
                    hs[j,i] = (f1-f2-f3+f4)/(4*eps*eps)
    
        return hs
